/*
 *  Copyright  1995-1997 AROS
 *  Copyright  1999-2010 MorphOS
 *
 *  $Id: liblist.c,v 1.3 2010/06/13 17:49:39 tokai Exp $
 */

#include <exec/memory.h>
#include <exec/tasks.h>
#include <exec/execbase.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <aros/machine.h>

#include "liblist_version.h"

static const char version[] = VERSTAG;

struct lib
{
	STRPTR name;
	APTR address;
	WORD version;
	WORD revision;
	WORD opencnt;
	UBYTE flags;
};


#define TEMPLATE "NAME"

enum
{
	ARG_NAME = 0,
	NUMARGS,
};


static int fillbuffer(struct lib **buffer, IPTR size, struct ExecBase *SysBase);

int main(void)
{
	struct ExecBase   *SysBase;
	struct DosLibrary *DOSBase;
	IPTR size;
	struct lib *buffer,*libs,*libs2;
	LONG error = RETURN_OK;

	SysBase = *((struct ExecBase**) 4);

	if ((DOSBase = (struct DosLibrary *)OpenLibrary("dos.library",37)) != 0)
	{
		struct RDArgs *rda;
		STRPTR argv[NUMARGS] = { 0 };

		if ((rda = ReadArgs(TEMPLATE, (LONG *)argv, NULL)))
		{
			STRPTR filter = argv[ARG_NAME] ? argv[ARG_NAME] : (STRPTR)"#?";
			ULONG  filterlen = strlen(filter) * 3 + 2;
			STRPTR filterp;

			if ((filterp = AllocVec(filterlen, MEMF_ANY)))
			{
				if (ParsePatternNoCase(filter, filterp, filterlen) != -1)
				{

					/*  Try to fill library list information into a 4K buffer; if not
					 *  enough try with a 2K larger buffer until everything fits.
					 */
					for(size=4096;;size+=2048)
					{
						buffer = AllocVec(size,MEMF_ANY);

						if(buffer==NULL)
						{
							error = RETURN_FAIL;
							SetIoErr(ERROR_NO_FREE_STORE);
							break;
						}

						libs = buffer;

						if (fillbuffer(&libs, size, SysBase))
						{
							FPuts(Output(),
								"Address     Version     Rev.  OpenCnt  Flags        Name\n"
								"----------  -------  -------  -------  -----------  -------------------------------\n");

							for(libs2=buffer; libs2<libs; libs2++)
							{
								IPTR args[12];

								if (!MatchPatternNoCase(filterp, libs2->name))
								{
									continue;
								}

								args[ 0] = (IPTR)libs2->address;
								args[ 1] = (IPTR)libs2->version;
								args[ 2] = (IPTR)libs2->revision;
								args[ 3] = (IPTR)libs2->opencnt;
								args[ 4] = (IPTR)((libs2->flags & LIBF_SUMMING)   ? 'S' : '-');
								args[ 5] = (IPTR)((libs2->flags & LIBF_CHANGED)   ? 'C' : '-');
								args[ 6] = (IPTR)((libs2->flags & LIBF_SUMUSED)   ? 'U' : '-');
								args[ 7] = (IPTR)((libs2->flags & LIBF_DELEXP )   ? 'D' : '-');
								args[ 8] = (IPTR)((libs2->flags & LIBF_RAMLIB)    ? 'R' : '-');
								args[ 9] = (IPTR)((libs2->flags & LIBF_QUERYINFO) ? 'Q' : '-');
								args[10] = (IPTR)libs2->flags;
								args[11] = (IPTR)libs2->name;

								VPrintf("0x%08.8lx  %7.7lu  %7.7lu  %7.7lu  %lc%lc%lc%lc%lc%lc 0x%02.2lx  %s\n", args);

								if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
								{
									error = RETURN_ERROR;
									SetIoErr(ERROR_BREAK);
									break;
								}
							}

							FreeVec(buffer);
							break; /* everything OK, leave for loop */
						}

						FreeVec(buffer);

					} /* for */
				}
				else
				{
					error = RETURN_FAIL;
				}

				FreeVec(filterp);
			}
			else
			{
				error = RETURN_FAIL;
				SetIoErr(ERROR_NO_FREE_STORE);
			} 

			FreeArgs(rda);
		}
		else
		{
			error = RETURN_ERROR;

		} /* ReadArgs */

		if (error != RETURN_OK)
		{
			PrintFault(IoErr(), "LibList");
		}

		CloseLibrary((struct Library *)DOSBase);
	}

	return error;
}


static int addlib(struct Library *lib, struct lib **l, STRPTR *e)
{
	STRPTR s1, s2;

	(*l)->address  = lib;
	(*l)->version  = lib->lib_Version;
	(*l)->revision = lib->lib_Revision;
	(*l)->opencnt  = lib->lib_OpenCnt;
	(*l)->flags    = lib->lib_Flags;

	s1 = lib->lib_Node.ln_Name;

	if(s1!=NULL)
	{
		s2=s1;

		while(*s2++)
			;

		while(s2>s1)
		{
			if(*e<=(STRPTR)(*l+2))
			{
				return 0;
			}

			*--(*e)=*--s2;
		}
	}

	if((STRPTR)(*l+2)>*e)
	{
		return 0;
	}

	(*l)->name=*e;
	++*l;

	return 1;
}


static int fillbuffer(struct lib **buffer, IPTR size, struct ExecBase *SysBase)
{
	STRPTR end = (STRPTR)*buffer+size;
	struct Library *lib;
	int ret = 1;
    
	Forbid();

	for(lib=(struct Library *)SysBase->LibList.lh_Head;
	    lib->lib_Node.ln_Succ!=NULL && ret;
	    lib=(struct Library *)lib->lib_Node.ln_Succ)
	{
		ret = addlib(lib,buffer,&end);
	}

	Permit();

	return ret;
}

